/*
	Model and learning algorithm for the tucker decomposition model.

	Notes:
	Such an approach has been proposed for context-aware recommendation in: A. Karatzoglou, X. Amatriain, L. Baltrunas, and N. Oliver (2010): Multiverse recommendation: n-dimensional tensor factorization for context-aware collaborative filtering, in Proceedings of the fourth ACM conference on Recommender systems (Recsys 2010)

	Author:   Steffen Rendle, http://www.libfm.org/
	modified: 2011-07-14

	Copyright 2010-2011 Steffen Rendle, see license.txt for more information
*/

#ifndef METHOD_TUCKER_H_
#define METHOD_TUCKER_H_

#include <cmath>
#include "Data.h"
#include "../../util/matrix.h"
#include "../../util/rlog.h"
#include "../../util/util.h"

// data structure to treat a double-vector as a m-mode tensor
class TensorIndex {
	private:		
		// The number of modes of the tensor		
		uint num_modes;
		// The number of dimensions of each mode in the tensor
		DVector<uint> num_dim;
		
		uint increment_mode;
	public:
		uint counter;
		DVector<uint> index;

		TensorIndex(uint num_modes, DVector<uint> & num_dim) {
			this->num_modes = num_modes;
			this->num_dim.assign(num_dim);
			for (uint m = 0; m < num_modes; m++) {
				assert(this->num_dim(m) > 0);
			}
			index.setSize(num_modes);
			first();
		}


		void first() {
			index.init(0);
			counter = 0;
			increment_mode = 0;
		}

		void next() {
			assert(increment_mode < num_modes);
			
			counter++;

			index(increment_mode)++;
			if (index(increment_mode) >= num_dim(increment_mode)) {
				do {
					increment_mode++;
					if (increment_mode >= num_modes) {
						return;
					}
					index(increment_mode)++;
				} while (index(increment_mode) >= num_dim(increment_mode));		
				for (uint m = 0; m < increment_mode; m++) {
					index(m) = 0;
				}
				increment_mode = 0;
			}	
		
		}

		bool valid_index() {
			return (increment_mode < num_modes); 
		}

		void print() {
			std::cout << counter << ":";
			for (uint m = 0; m < num_modes; m++) {
				std::cout << "\t" << index(m);
			}
			std::cout << std::endl;
		}
};


class TuckerTensorFactorization {
	protected:
		uint num_modes;
		
		DVector< DMatrixDouble > factors;
		DVectorDouble core;

		DVector< DVectorDouble > grad_factors;
		DVectorDouble grad_core;
		
	public:
		RLog* log;

		double bias;
		
		DVector< uint > k;
		double regular_factors;
		double regular_core;
		
		double init_stdev;
		double learn_rate;
		double decay;
		uint num_iterations;

		TuckerTensorFactorization() { log = NULL; } 

		void init(DVector<Data*>& data_source) {
			// get dimensions of the data set
			DVector<uint> max_index;
			{
				assert(data_source.dim > 0);
				num_modes = data_source(0)->data.dim2 - 1;
				std::cout << "num_modes=" << num_modes << std::endl;
				max_index.setSize(num_modes);
				max_index.init(0); 
				for (uint d = 0; d < data_source.dim; d++) {
					assert(data_source(d)->data.dim2 == (num_modes+1));
					for (uint i = 0; i < data_source(d)->data.dim1; i++) {
						for (uint m = 0; m < num_modes; m++) {
							assert(data_source(d)->data(i,m) >= 0);
							max_index(m) = std::max(max_index(m), (uint) data_source(d)->data(i,m));
						}
					}
				}
			}
				
			// debug info:
			std::cout << "dim of mode:\t(";
			for (uint m = 0; m < num_modes; m++) {
				std::cout << max_index(m)+1 << ",";
			}		
			std::cout << ")" << std::endl;std::cout.flush();


			assert(k.dim == num_modes);
			std::cout << "dim of factorization:\t(";
			for (uint m = 0; m < num_modes; m++) {
				std::cout << k(m) << ",";
			}		
			std::cout << ")" << std::endl;std::cout.flush();
			

			// get the number of factors in the core tensor
			uint num_core_factors = 1;
			{
				assert(num_modes > 0);
				for (uint m = 0; m < num_modes; m++) {
					num_core_factors *= k(m);
				}
			}
			std::cout << "num_core_factors=" << num_core_factors << std::endl;
	
			

			// create the data structures for core, factors and their gradients:
			core.setSize(num_core_factors);
			core.init_normal(0.0, init_stdev);
			grad_core.setSize(num_core_factors);

			factors.setSize(num_modes);
			grad_factors.setSize(num_modes);
			for (uint m = 0; m < num_modes; m++) {	
				factors(m).setSize(max_index(m)+1, k(m));
				factors(m).init(0.0, init_stdev);
				grad_factors(m).setSize(k(m));
			}
			bias = 0;

			// check the index
			if (false) {
				TensorIndex idx (num_modes, k);
				while (idx.valid_index()) {
					idx.print();
					idx.next();
				}
				std::cout << "finished index test" << std::endl;
				std::cout.flush();				
			}
			std::cout << "init has finished" << std::endl;
			std::cout.flush();

			if (log != NULL) {
				log->addField("rmse", std::numeric_limits<double>::quiet_NaN());
				log->addField("mae", std::numeric_limits<double>::quiet_NaN());
				log->addField("time_pred", std::numeric_limits<double>::quiet_NaN());
				log->addField("time_learn", std::numeric_limits<double>::quiet_NaN());
			}				
			
		}

		double predict(int* index) {
			double result = bias;
			double v;
			TensorIndex idx (num_modes, k);
			while (idx.valid_index()) {
				v = core(idx.counter);
				for (uint m = 0; m < num_modes; m++) {
					v *= factors(m)(index[m], idx.index(m));
				}
				result += v;
				idx.next();
			}
			return result;
		}

		void train(Data& train, Data& test) {
			// debug info:
			std::cout << "regular_core=" << regular_core << std::endl;
			std::cout << "regular_factors=" << regular_factors << std::endl;
			std::cout << "learn_rate=" << learn_rate << std::endl;
			std::cout << "decay=" << decay << std::endl;
			std::cout << "init_stdev=" << init_stdev << std::endl;
			std::cout << "num_iterations=" << num_iterations << std::endl;
			std::cout.flush();

			// set the bias
			long long sum = 0;
			for (uint j = 0; j < train.data.dim1; j++) {
				sum += train.target(j);
			}
			bias = (double) sum / train.data.dim1; 

			// train
			for (uint i = 0; i < num_iterations; i++) {
				
				double iteration_time = getusertime();
				for (uint j = 0; j < train.data.dim1; j++) {
					SGD_learn_from_case(train.data(j), train.target(j));
				}	
				iteration_time = (getusertime() - iteration_time);
				if (log != NULL) {
					log->log("time_learn", iteration_time);
				}
				evaluate(test); std::cout.flush();	
				if (log != NULL) {
					log->newLine();
				}
				learn_rate *= decay;
			}
		}

		void evaluate(Data& test) {
			double mae = 0.0;
			double rmse = 0.0;
			double eval_time = getusertime();
			for (uint i = 0; i < test.data.dim1; i++) {
				double error = predict(test.data(i)) - test.target(i);
				rmse += error*error;
				mae += std::abs((double)error);	
			}
			mae /= test.data.dim1;
			rmse = std::sqrt(rmse / test.data.dim1);

			eval_time = (getusertime() - eval_time);
			// log the values
			if (log != NULL) {
				log->log("rmse", rmse);
				log->log("mae", mae);
				log->log("time_pred", eval_time);
			}

			std::cout << "RMSE/ MAE:\t" << rmse << "\t" << mae << std::endl;
		}

		void SGD_learn_from_case(int* index, double target) {
			double pred = predict(index);

			double grad_direction;
			if (true) {
				grad_direction = 2 * (pred-target);
			} else {
				if (pred == target) { 
					grad_direction = 0;
				} else if (pred > target) {
					grad_direction = 1;
				} else {
					grad_direction = -1;
				}
			}

			// calculate gradient of the core:
			{
				grad_core.init(0.0);
				double v;
				TensorIndex idx (num_modes, k);
				while (idx.valid_index()) {
					v = 1.0;
					for (uint m = 0; m < num_modes; m++) {
						v *= factors(m)(index[m], idx.index(m));
					}
					grad_core(idx.counter) = v;
					idx.next();
				}
			}			

			// calculate gradient for each factor vector of i_j:
			{
				for (uint m = 0; m < num_modes; m++) {
					grad_factors(m).init(0.0);
				}
				double v;
				TensorIndex idx (num_modes, k);
				while (idx.valid_index()) {
					for (uint m_o = 0; m_o < num_modes; m_o++) {
						v = 1.0;
						for (uint m = 0; m < num_modes; m++) {
							if (m != m_o) {
								v *= factors(m)(index[m], idx.index(m));
							}
						}
						v *= core(idx.counter);
						grad_factors(m_o)(idx.index(m_o)) += v;		
					}
					idx.next();
				}
			}			

			// apply SGD:
			{
				// on core:
				TensorIndex idx (num_modes, k);
				while (idx.valid_index()) {
					core(idx.counter) -= learn_rate * ( grad_direction * grad_core(idx.counter) + regular_core * core(idx.counter) );
					idx.next();
				}

				// on feature matrices:
				for (uint m = 0; m < num_modes; m++) {
					for (uint f = 0; f < k(m); f++) {
						factors(m)(index[m], f) -= learn_rate * ( grad_direction * grad_factors(m)(f) + regular_factors * factors(m)(index[m], f) );
					}
				}
			}		

		}	
		
		
	
};

#endif /*METHOD_TUCKER_H_*/
