/*
	Tensor factorization for regression based on the Tucker decomposition model.

	Notes:
	Such an approach has been proposed for context-aware recommendation in: A. Karatzoglou, X. Amatriain, L. Baltrunas, and N. Oliver (2010): Multiverse recommendation: n-dimensional tensor factorization for context-aware collaborative filtering, in Proceedings of the fourth ACM conference on Recommender systems (Recsys 2010)

	Author:   Steffen Rendle, http://www.libfm.org/
	modified: 2011-07-14

	Copyright 2010-2011 Steffen Rendle, see license.txt for more information
*/


#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <string>
#include <iterator>
#include <algorithm>
#include <iomanip>
#include "../util/util.h"
#include "../util/cmdline.h"

#include "src/tucker.h"

using namespace std;

int main(int argc, char **argv) { 
 	
	srand ( time(NULL) );
	try {
		CMDLine cmdline(argc, argv);
		std::cout << "Tensor factorization" << std::endl;
		std::cout << "  Version: 1.00" << std::endl;
		std::cout << "  Author:  Steffen Rendle, srendle@ismll.de, http://www.libfm.org/" << std::endl;
		std::cout << "  License: Free for academic use. See license.txt." << std::endl;
		std::cout << "----------------------------------------------------------------------------" << std::endl;

	

		// Input/ output files:
		const std::string param_train_file	= cmdline.registerParameter("train", "filename for training data [MANDATORY]");
		const std::string param_test_file	= cmdline.registerParameter("test", "filename for test data [MANDATORY]");
		const std::string param_out		= cmdline.registerParameter("out", "filename for output; default=''");
	

		// Model parameters: 
		const std::string param_dim		= cmdline.registerParameter("dim", "dim of factorization; default=8");
	
		// Optimization parameters:
		const std::string param_regular		= cmdline.registerParameter("regular", "regularization (matrices;core); default=0.0;0.0");


		// Learning parameters:
		const std::string param_init_stdev	= cmdline.registerParameter("init_stdev", "standard deviation for factor initialization; default=0.1");			
		const std::string param_num_iter	= cmdline.registerParameter("iter", "number of iterations for SGD; default=100");
		const std::string param_learn_rate	= cmdline.registerParameter("learn_rate", "learn_rate for SGD; default=0.1");
		const std::string param_decay      	= cmdline.registerParameter("decay", "decay in learning rate, default=1.0");

		// program parameters:
		const std::string param_verbosity	= cmdline.registerParameter("verbosity", "how much infos to print; default=0");
		const std::string param_r_log		= cmdline.registerParameter("rlog", "write measurements within iterations to a file; default=''");
		const std::string param_help            = cmdline.registerParameter("help", "this screen");

		if (cmdline.hasParameter(param_help) || (argc == 1)) {
			cmdline.print_help();
			return 0;
		}
		cmdline.checkParameters();		


		// (1) Load the data
		std::cout << "Loading train...\t";
		Data train;
		train.load(cmdline.getValue(param_train_file));
		if (cmdline.getValue(param_verbosity, 0) > 0) { train.debug(); }
		std::cout << "Loading test... \t";
		Data test;
		test.load(cmdline.getValue(param_test_file));
		if (cmdline.getValue(param_verbosity, 0) > 0) { test.debug(); }


		ofstream* out_rlog = NULL;
	 	if (cmdline.getValue(param_r_log).compare("")) {
	 		out_rlog = new ofstream(cmdline.getValue(param_r_log).c_str());
	 		if (! out_rlog->is_open())	{
	 			throw "Unable to open file " + cmdline.getValue(param_r_log);
	 		}
	 		std::cout << "logging to " << cmdline.getValue(param_r_log).c_str() << std::endl;
	 	}
	 	RLog rlog(out_rlog);

		// (2) Setup the model
		TuckerTensorFactorization pred;
		{
			{ 
				vector<int> dim = cmdline.getIntValues(param_dim);
				assert(dim.size() > 0);
				pred.k.setSize(dim.size());
				for (uint i = 0; i < dim.size(); i++) {
					pred.k(i) = dim[i];
				}				
			}

			{
				vector<double> reg = cmdline.getDblValues(param_regular);
				assert(reg.size() == 2);
				pred.regular_factors = reg[0];
				pred.regular_core = reg[1];
			}							
										
			pred.learn_rate = cmdline.getValue(param_learn_rate, 0.1);
			pred.decay = cmdline.getValue(param_decay, 1.0);
			pred.init_stdev = cmdline.getValue(param_init_stdev, 0.1);
			pred.num_iterations = cmdline.getValue(param_num_iter, 100);
			pred.log = &rlog;
		
			// init the model:
			{
				DVector<Data*> data;
				data.setSize(2);
				data(0) = &train;
				data(1) = &test;
				pred.init(data);
				rlog.init();
			}

			// train:
			pred.train(train, test);

		}
						 	
	
	} catch (std::string &e) {
		std::cerr << e << std::endl;
	} catch (char const* &e) {
		std::cerr << e << std::endl;
	}

}

