/*
	Dense Matrix and Vectors

	Author:   Steffen Rendle, http://www.libfm.org/
	modified: 2011-07-18

	Copyright 2010-2011 Steffen Rendle, see license.txt for more information
*/

#ifndef MATRIX_H_
#define MATRIX_H_

#include <vector>
#include <assert.h>
#include <iostream>
#include <fstream>
#include "../util/random.h"
#include "../util/token_reader.h"


template <typename T> class DMatrix {
	public:
		std::vector<std::string> col_names;
		uint dim1, dim2;
		T** value;
		
		T get(uint x, uint y) {
   			assert((x < dim1) && (y < dim2));
			return value[x][y];
		}

		DMatrix(uint p_dim1, uint p_dim2) {
			dim1 = 0;
			dim2 = 0;	
			value = NULL;
			setSize(p_dim1, p_dim2);
		}
		
		DMatrix() {
			dim1 = 0;
			dim2 = 0;	
			value = NULL;
		}
		
		~DMatrix() {
			if (value != NULL) {
				delete [] value[0];
				delete [] value;
			}	
		}
		
		void setSize(uint p_dim1, uint p_dim2) {
			if (value != NULL) {
				delete [] value[0];
				delete [] value;
			}
			dim1 = p_dim1;
			dim2 = p_dim2;
			value = new T*[dim1];
			value[0] = new T[dim1 * dim2];
			for (unsigned i = 1; i < dim1; i++) {
				value[i] = value[0] + i * dim2;
			}			
			col_names.resize(dim2);
			for (unsigned i = 1; i < dim2; i++) {
				col_names[i] = "";
			}						
		}
		
		T& operator() (unsigned x, unsigned y) {
   			assert((x < dim1) && (y < dim2));
			return value[x][y];	
		}
   		T operator() (unsigned x, unsigned y) const {
   			assert((x < dim1) && (y < dim2));
   			return value[x][y];
   		}
   		
   		T* operator() (unsigned x) const {
   			assert((x < dim1));
   			return value[x];
   		}
   		
   		void save(std::string filename, bool has_header = false) {
		   	std::ofstream out_file (filename.c_str());
			if (out_file.is_open())	{
				if (has_header) {
					for (uint i_2 = 0; i_2 < dim2; i_2++) {
						if (i_2 > 0) {
							out_file << "\t";
						}
						out_file << col_names[i_2];					
					}	
					out_file << std::endl;
				}
				for (uint i_1 = 0; i_1 < dim1; i_1++) {
					for (uint i_2 = 0; i_2 < dim2; i_2++) {
						if (i_2 > 0) {
							out_file << "\t";
						}
						out_file << value[i_1][i_2];
					}
					out_file << std::endl;
				}
				out_file.close();
			} else {
				std::cout << "Unable to open file " << filename;
			}   			
   		}
   		
   		void load(std::string filename) {
			std::ifstream in_file (filename.c_str());
		  	if (in_file.is_open()) {
				token_reader fData2(&in_file);
				for (uint i_1 = 0; i_1 < dim1; i_1++) {
					for (uint i_2 = 0; i_2 < dim2; i_2++) {
						T v;
						fData2.read(v);
						value[i_1][i_2] = v;
					}
				}
				in_file.close();
			} else {
				throw "Unable to open file " + filename;
			}		
		}

		void _load(std::string filename, bool has_header) {//, vector<std::string>& header) {
			// (1) get the dimensionality
			int num_lines = 0;
			int num_values_per_line = 0;
			int num_values = -1;
			{
				std::ifstream fData(filename.c_str());
			 	if (! fData.is_open()) {
			 		std::cout << "unable to open " << filename << std::endl;
			 		return;
			 	}
				token_reader fData2(&fData);
				bool do_cnt_values_per_line = true;
				int num_values_in_current_line = 0;
				do {
					int i =	fData2.skipValue();
					num_values++;
					num_values_in_current_line++;
					if (i == 0) {
						break;
					} else if (i == 1) {
						if (do_cnt_values_per_line) {
							do_cnt_values_per_line = false;
							num_values_per_line = num_values_in_current_line;
						} else {
							assert(num_values_per_line == num_values_in_current_line);
						}
						num_values_in_current_line = 0;
						num_lines++;
					}
				} while (true);
				if (has_header) {
					num_lines--;
					num_values -= num_values_per_line;
				}	
			}		
			
			
			// (2) assign the size
			std::cout << "number of columns: " << num_values_per_line << std::endl;
			std::cout << "number of lines:   " << num_lines << std::endl;
			std::cout << "number of values:  " << num_values << std::endl; 
			setSize(num_lines, num_values_per_line);
			
			// (3) read the data
			{
				std::ifstream fData(filename.c_str());
			 	if (! fData.is_open()) {
			 		std::cout << "unable to open " << filename << std::endl;
			 		return;
			 	}
				token_reader fData2(&fData);
				if (has_header) {
					for (uint i2 = 0; i2 < dim2; i2++) {
						col_names[i2] = fData2.readString();
					}
				}
				for (uint x = 0; x < dim1; x++) {
					for (uint y = 0; y < dim2; y++) {
						value[x][y] = fData2.readInt();
					}
				}
			}
		}
   		

};

template <typename T> class DVector {
	public:
		uint dim;
		T* value;
		DVector() {
			dim = 0;
			value = NULL;
		}
		DVector(uint p_dim) {
			dim = 0;
			value = NULL;
			setSize(p_dim);
		}
		~DVector() {
			if (value != NULL) {
				delete [] value;
			}	
		}
		T get(uint x) {
			return value[x];
		}
		void setSize(uint p_dim) {
			if (value != NULL) {
				delete [] value;
			}
			dim = p_dim;
			value = new T[dim];			
		}
		T& operator() (unsigned x) {
			return value[x];	
		}
   		T operator() (unsigned x) const {
   			return value[x];
   		}
   		void init(T v) {
   			for (uint i = 0; i < dim; i++) {
   				value[i] = v;
   			}	
   		}
   		void assign(T* v) {
			if (v->dim != dim) { setSize(v->dim); }
   			for (uint i = 0; i < dim; i++) {
   				value[i] = v[i];
   			}
   		}
		void assign(DVector<T>& v) {
			if (v.dim != dim) { setSize(v.dim); }
   			for (uint i = 0; i < dim; i++) {
   				value[i] = v.value[i];
   			}
   		}
   		void save(std::string filename) {
		   	std::ofstream out_file (filename.c_str());
			if (out_file.is_open())	{
				for (uint i = 0; i < dim; i++) {
					out_file << value[i] << std::endl;
				}
				out_file.close();
			} else {
				std::cout << "Unable to open file " << filename;
			}   			
   		}
};


class DVectorDouble : public DVector<double> {
	public:
		void init_normal(double mean, double stdev) {	
			for (uint i_2 = 0; i_2 < dim; i_2++) {
				value[i_2] = ran_gaussian(mean, stdev);
			}
		}
};

class DMatrixDouble : public DMatrix<double> {
	public:
		void init(double mean, double stdev) {	
			for (uint i_1 = 0; i_1 < dim1; i_1++) {
				for (uint i_2 = 0; i_2 < dim2; i_2++) {
					value[i_1][i_2] = ran_gaussian(mean, stdev);
				}
			}
		}
		void init_column(double mean, double stdev, int column) {	
			for (uint i_1 = 0; i_1 < dim1; i_1++) {
				value[i_1][column] = ran_gaussian(mean, stdev);
			}
		}
};


#endif /*MATRIX_H_*/
